c	**************************************************************
c	Routine to determine additional geometry parameters required
c	to calculate the frequency scale in the radiative transfer
c	computations
c	**************************************************************
	subroutine initfreq(vz,vt,nx,ny,nz,dny,nff,dnyred,nffred,kerr)
	implicit none
	include 'fsizes.inc'

c	input parameter
	integer nx,ny,nz
	real vt(ns,ns,ns),vz(ns,ns,ns)

c	output parameter
	real dny,dnyred,vmax
	integer nff,nffred,kerr
	
c	internal variables
	real negexp,negint,dvgauss,epsn
	real dvmin
	integer npoi,i,j,k
	common /accgrid/ epsn,dvgauss
	common /accgauss/ negexp,negint
	save /accgrid/,/accgauss/

c	Points to cover half the frequency range
	data npoi / 25 /

c	***************************************************************
c	Find the minimum line width
c	Find the extreme systematic velocities
c	***************************************************************
	kerr=0
	dvmin=1e30
	vmax=0.0
	do 100 i=1,nx
	do 100 j=1,ny
	do 100 k=1,nz
	dvmin=min(dvmin,vt(i,j,k))
	vmax=max(vmax,abs(vz(i,j,k))+vt(i,j,k)*negint)
100	continue
	dnyred=dvmin*dvgauss
	nffred=int(vmax/dnyred)+1
c	**************************************************************
c	Reduce number of frequency points for RT iteration
c	**************************************************************
	if (nffred.gt.npoi) then
	nffred=npoi
	dnyred=vmax/npoi
	endif
c	**************************************************************
c	Extreme velocities - check whether nf is sufficient
c	**************************************************************
	if (nffred.gt.nf) kerr=1
	nff=int(vmax/dny)+1
	if (nff.gt.nf) kerr=1
	return
	end

c	**************************************************************
c	First guess for the excitation ratios 
c	For a table of density/temperature values only
c	**************************************************************
	subroutine ddtable(dh,t,uext,cptb,sctb,lin,lev,
     %                        denstab,momtab,regtab,ntt,inum)
	implicit none
	integer inum,lin,lev,itmx,ntt,direct

	include 'fsizes.inc'
	parameter (itmx=100,direct=6)
	real*8 negvek(nlev)
	real uext(ntab,ntra),sctb(ntab,ntra),cptb(ntab,ntra)
	real denstab(ndens),momtab(ntab,6),grd(direct)
	real regtab(ntab,6),dnsgrd(direct)
	real dh,t

	real*8 dd(nlev),sumdn,probtherm
	real cp(ntra),sc(ntra),ubg(ntra),hdens
	integer i,j,l,itermax,imx,ndstep,index

c	**********************************************************
c	Internal formats
c	**********************************************************
11	format(/,' LVG did not reach convergence at density: ',
     %            E11.4,',gradient: ',I4)
12	format(' Maximum number of iterations in Sobolev ',
     %  'approximation: ',I3)
13	format(' Required number of rotational levels: ',I2)
c	************************************************************
c	Array for level reduction
c	************************************************************
	do 110 l=1,lev
110	negvek(l)=1d-14
c	************************************************************
c	Compute the level populations for all densities depending on 
c	the initialization parameter
c	************************************************************
	if (abs(inum).eq.1) then
c	************************************************************
c	Thermal population
c	************************************************************
	sumdn=0.0d0
	do 150 l=1,lev
	dd(l)=probtherm(t,l)
	sumdn=sumdn+dd(l)
150	continue
	do 180 l=1,lev
	 dd(l)=dd(l)/sumdn
	 if (dd(l).gt.negvek(l)) negvek(l)=dd(l)
180	continue
	call kapsrc(dd,cp,sc,lin)
	do 300 i=1,ntt
	do 300 l=1,lin
	cptb(i,l)=cp(l)
	sctb(i,l)=sc(l)
300	continue
	else if (abs(inum).eq.2) then
c	************************************************************
c	Only cosmic background as external radiation
c	************************************************************
	do 400 i=1,ntt
	hdens=denstab(i)*dh
	do 410 l=1,lin
410	ubg(l)=uext(i,l)
	call matrixup(t,hdens,ubg,dd,lin,lev)
	do 450 l=1,lev
	 if (dd(l).gt.negvek(l)) negvek(l)=dd(l)
450	continue
	call kapsrc(dd,cp,sc,lin)
	do 400 l=1,lin
	  cptb(i,l)=cp(l)
	  sctb(i,l)=sc(l)
400	continue
	else
c	************************************************************
c	Sobolev program 
c	Cosmic background solution as initial guess
c	************************************************************
	itermax=1
	ndstep=ntt/ngrad
	do i=1,ndstep
	hdens=denstab(i)*dh
	do j=ngrad,1,-1
	index=(i-1)*ngrad+j
	do 510 l=1,lin
510	ubg(l)=uext(index,l)
c	if (j.eq.ngrad) call matrixup(t,hdens,ubg,dd,lin,lev)
	call matrixup(t,hdens,ubg,dd,lin,lev)
	imx=itmx
	call dirfromdst3(momtab,index,grd)
	call bgrad(regtab,index,grd,dnsgrd)
	call sobolev(hdens,t,grd,dnsgrd,ubg,lin,dd,lev,imx)
	if (imx.ge.itmx) write(6,11) denstab(i),j
	itermax=max(imx,itermax)
	do 580 l=1,lev
	 if (dd(l).gt.negvek(l)) negvek(l)=dd(l)
580	continue
	call kapsrc(dd,cp,sc,lin)
	do l=1,lin
	 cptb(index,l)=cp(l)
	 sctb(index,l)=sc(l)
	enddo
	enddo
	enddo
	write(6,12) itermax
	endif
	call removelevels(negvek,lev)
	write(6,13) lev
	return
	end

c	**************************************************************
c	Small test routine to write out gradients
c	**************************************************************
	subroutine printgradients(i,j,inx,denstab,momtab,regtab)
	implicit none
	integer i,j,direct

	include 'fsizes.inc'
	parameter (direct=6)
	real denstab(ndens),momtab(ntab,6),grd(direct)
	real regtab(ntab,6),dnsgrd(direct)
	real dh,t
	integer inx(ns,ns,ns),index,k,l

	do k=1,128
	  index=inx(i,j,k)
	  call dirfromdst3(momtab,index,grd)
	  call bgrad(regtab,index,grd,dnsgrd)
	  write(6,*) i,j,k,denstab((index-1)/ngrad+1)
	  write(6,*) grd,dnsgrd
	enddo
	return
	end

c	**************************************************************
c	Make table of density/temperature values 
c	**************************************************************
c	Preparation for Sobolev approximation and frequency
c	integration
c	Compute distributions of gradients and velocities for 
c	each density table point
c	**********************************************************
	subroutine maketables(dco,vx,vy,vz,vt,dinx,nx,ny,nz,dr,dny,
     %                     denstab,gradfct,densmom,freqmom,ntt,inum)
	implicit none
	integer ntt,nx,ny,nz,inum

	include 'fsizes.inc'
	real dco(ns,ns,ns),vz(ns,ns,ns),vt(ns,ns,ns),denstab(ndens)
	real vx(ns,ns,ns),vy(ns,ns,ns)
	real dr,dny,dcomax,dcomin,step,dvmin

	real dval,vgrad,grad,freq,sigcut,hlp,sqabs,vxshift,vyshift
	real freqmom(ntab,2),gradfct(ntab,6),densmom(ntab,6)
	real gradmom(ndens,2),grad6(6),dgrad(6)
	real*8 delta,sumgrd(ntab,6),sumqgrd(ntab,6)
	real*8 sumtherm(ntab),sumv(ntab),sumqv(ntab)
	real*8 sumf6(ntab,6),vxav,vyav,vzav,dtot
	integer dinx(ns,ns,ns)
	integer dirtab(ntab),shifttab(ndens),order(6)
	integer i,j,k,l,index,nindex
	integer ineu,ilast,inext,jlast,jnext,klast,knext

	real epsn,dvgauss
	common /accgrid/ epsn,dvgauss
	save /accgrid/

c	90% of the nomral distribution coverd at 1.16 sigma
	data sigcut /1.16/
c	translation between the average absolute value and sigma
	data sqabs /0.5/
c	Exponential distribution
c	data sqabs /0.5/
c	Normal distribution
c	data sqabs /0.797885/

c	**********************************************************
c	Internal formats
c	**********************************************************
13	format(/,' Number of density points: ',I3)
12	format(' Number of density points reduced to: ',I3)
14	format(' Total number of grid points: ',I5)
c	**************************************************************
c	The excitation rates are computed for a table of densities
c	and temperatures only. First step - only densities
c	**************************************************************
c	First run through the cube
c	Find the size of the density table: max+min density
c	**************************************************************
	dcomax=0.0
	dcomin=1e30
	vxav=0.0d0
	vyav=0.0d0
	vzav=0.0d0
	dtot=0.0d0
	do 100 i=1,nx
	do 100 j=1,ny
	do 100 k=1,nz
	 dcomax=max(dcomax,dco(i,j,k))
	 dcomin=min(dcomin,dco(i,j,k))
	 dtot=dtot+dco(i,j,k)
	 vxav=vxav+dco(i,j,k)*vx(i,j,k)
	 vyav=vyav+dco(i,j,k)*vy(i,j,k)
	 vzav=vzav+dco(i,j,k)*vz(i,j,k)
100	continue
c	Use extrema and construct table
	dcomin=max(dcomin,1e-9*dcomax)
	step=alog(epsn)
	ntt=int(alog(dcomax/dcomin)/step)+1
	if (ntt.gt.ntab) ntt=ntab
	write(6,13) ntt
	dcomin=dcomax*exp(step*(1-ntt))
c	***********************************************************
c	Compute the global velocity centroids in each direction to
c	allow a shift of the directional velocities relative to 0
c	***********************************************************
	vxav=vxav/dtot
	vyav=vyav/dtot
	vzav=vzav/dtot
	vxshift=vxav-vzav
	vyshift=vyav-vzav
c	***********************************************************
c	Now compute the distributions of gradients
c	Initialize the accumulators for the distributions
c	***********************************************************
	do 220 i=1,ntt
	dirtab(i)=0
	sumv(i)=0.0d0
        sumqv(i)=0.0d0
220	continue
c	***********************************************************
c	Scan the field 
c	second run - get the global gradient distribution
c	For this rough statistics the z component is taken to be
c	sufficient
c	***********************************************************
	do k=1,nz
	 klast=mod(k+nz-2,nz)+1
	 knext=mod(k,nz)+1
	do i=1,nx
	do j=1,ny
c	Find the closest density values from the table
	  if (dco(i,j,k).le.dcomin) then
	    index=ntt
	  else
	    index=nint(alog(dcomax/dco(i,j,k))/step)+1
	  endif
	  dinx(i,j,k)=index
c	Now accumulate the gradients
	  dirtab(index)=dirtab(index)+2
	  dval=max(dco(i,j,k),dcomin)
	  dvmin=vt(i,j,k)/(nz*0.15)
c	Velocity gradient 1
	  vgrad=vz(i,j,k)-vz(i,j,klast)
	  grad=alog(max(abs(vgrad),dvmin)/(dval*dr))
	  sumv(index)=sumv(index)+grad
	  sumqv(index)=sumqv(index)+grad**2
c	Velocity gradient 2
	  vgrad=vz(i,j,k)-vz(i,j,knext)
	  grad=alog(max(abs(vgrad),dvmin)/(dval*dr))
	  sumv(index)=sumv(index)+grad
	  sumqv(index)=sumqv(index)+grad**2
	enddo
	enddo
	enddo
c	**********************************************************
c	Remove unused table entries - Compute moments
c	**********************************************************
	ineu=0
	do i=1,ntt
	 if (dirtab(i).gt.0) then
	  ineu=ineu+1
	  denstab(ineu)=dcomax*exp(step*(1-i))
	  gradmom(ineu,1)=sumv(i)/dirtab(i)
	  delta=sumqv(i)*dirtab(i)-sumv(i)**2
	  if (abs(delta).lt.0.0001*sumv(i)**2) then 
	   gradmom(ineu,2)=2.0
	  else
	   gradmom(ineu,2)=sqrt(2.0*delta)/dirtab(i)
	  endif
	  gradmom(ineu,2)=sigcut*gradmom(ineu,2)
	 endif
	 shifttab(i)=ineu
	enddo
	if (ineu.ne.ntt) then
	  ntt=ineu
	  write(6,12) ntt
	endif
c	***********************************************************
c	Now compute the distributions for each density-gradient pair
c	Initialize the new accumulators for the distributions
c	***********************************************************
	if (abs(inum).gt.2) then
	ntt=ntt*ngrad
	write(6,14) ntt
	do i=1,ntt
	  dirtab(i)=0
	  sumv(i)=0.0d0
          sumqv(i)=0.0d0
	  sumtherm(i)=0.0d0
	  do l=1,6
	    sumf6(i,l)=0.0d0
	    sumgrd(i,l)=0.0d0
	    sumqgrd(i,l)=0.0d0
	  enddo
	enddo
c	*********************************************************
c	Now construct the real table, assign cube elements
c	The decision is based on the average velocity gradient
c	Assign the points to the density fields
c	Compute the local velocity distributions
c	*********************************************************
	do k=1,nz
	 klast=mod(k+nz-2,nz)+1
	 knext=mod(k,nz)+1
	do i=1,nx
	 ilast=mod(i+nx-2,nz)+1
         inext=mod(i,nx)+1
	do j=1,ny
	 jlast=mod(j+ny-2,nz)+1
         jnext=mod(j,ny)+1
	  index=shifttab(dinx(i,j,k))
c	Compute directional gradients
	  dval=max(dco(i,j,k),dcomin)
	  dvmin=vt(i,j,k)/(nz*0.15)
	  hlp=1.0/(dval*dr)
	  grad6(1)=max(abs(vz(i,j,k)-vz(i,j,klast)),dvmin)*hlp
	  grad6(2)=max(abs(vz(i,j,k)-vz(i,j,knext)),dvmin)*hlp
	  grad6(3)=max(abs(vx(i,j,k)-vx(ilast,j,k)),dvmin)*hlp
	  grad6(4)=max(abs(vx(i,j,k)-vx(inext,j,k)),dvmin)*hlp
	  grad6(5)=max(abs(vy(i,j,k)-vy(i,jlast,k)),dvmin)*hlp
	  grad6(6)=max(abs(vy(i,j,k)-vy(i,jnext,k)),dvmin)*hlp
c	Density gradients
	  hlp=vt(i,j,k)*hlp/dval
	  dgrad(1)=-(dco(i,j,k)-dco(i,j,klast))*hlp
	  dgrad(2)=-(dco(i,j,k)-dco(i,j,knext))*hlp
	  dgrad(3)=-(dco(i,j,k)-dco(ilast,j,k))*hlp
	  dgrad(4)=-(dco(i,j,k)-dco(inext,j,k))*hlp
	  dgrad(5)=-(dco(i,j,k)-dco(i,jlast,k))*hlp
	  dgrad(6)=-(dco(i,j,k)-dco(i,jnext,k))*hlp
c	Compute average gradient
	  grad=0.0
	  do l=1,6
	    grad=grad+alog(grad6(l))
	  enddo
	  grad=grad/6.0
	  nindex=int((grad-gradmom(index,1))*0.5*ngrad
     %               /(gradmom(index,2)-gradmom(index,1))+0.5*ngrad)
	  nindex=min(max(nindex+1,1),ngrad)+ngrad*(index-1)
	  dinx(i,j,k)=nindex
	  dirtab(nindex)=dirtab(nindex)+1
c	Velocity gradients and density gradients (B factor from the Sobolev paper)
	  call gradorder(grad6,order)
	  do l=1,6
	    sumf6(nindex,l)=sumf6(nindex,l)+alog(grad6(order(l)))
	    sumgrd(nindex,l)=sumgrd(nindex,l)+dgrad(order(l))
	    sumqgrd(nindex,l)=sumqgrd(nindex,l)+dgrad(order(l))**2
	  enddo
c	Velocities  
	  freq=vz(i,j,k)/dny
	  sumv(nindex)=sumv(nindex)+freq
	  sumqv(nindex)=sumqv(nindex)+freq**2
	  freq=(vx(i,j,k)-vxshift)/dny
	  sumv(nindex)=sumv(nindex)+freq
	  sumqv(nindex)=sumqv(nindex)+freq**2
	  freq=(vy(i,j,k)-vyshift)/dny
	  sumv(nindex)=sumv(nindex)+freq
	  sumqv(nindex)=sumqv(nindex)+freq**2
	  sumtherm(nindex)=sumtherm(nindex)+vt(i,j,k)
	enddo
	enddo
	enddo
c	**********************************************************
c	Compute moments
c	**********************************************************
	call removezeros(sumf6,dirtab,ntt)
	do i=1,ntt
c	Points which do not occur are just set to fixed values
	  index=(i-1)/ngrad+1
	  if (dirtab(i).lt.1) then
	    freqmom(i,1)=0.0
	    freqmom(i,2)=1.0
	    k=mod(i-1,ngrad)+1
	    gradfct(i,1)=exp(gradmom(index,1))
	    gradfct(i,2)=-gradfct(i,1)
	    gradfct(i,3)=gradfct(i,1)
	    gradfct(i,4)=-gradfct(i,1)
	    gradfct(i,5)=gradfct(i,1)
	    gradfct(i,6)=-gradfct(i,1)
	  else
	    freqmom(i,1)=sumv(i)/(3*dirtab(i))
c	What is the better approach (makes only small differences) ?
	    freqmom(i,2)=sqrt(2.0*(sumqv(i)/(3*dirtab(i))
     %        -freqmom(i,1)**2)+(sumtherm(i)/(dny*dirtab(i)))**2)
c	    freqmom(i,2)=sumtherm(i)/(dny*dirtab(i))
	    do l=1,6
	     gradfct(i,l)=exp(sumf6(i,l)/dirtab(i))
	     densmom(i,l)=(sumgrd(i,l)+sqabs*(-1)**l
     %        *sqrt(sumqgrd(i,l)*dirtab(i)-sumgrd(i,l)**2))/dirtab(i)
	    enddo
	  endif
	enddo
c	***********************************************************
c	No gradients needed
c	***********************************************************
	else
	do i=1,ntt
	  dirtab(i)=0
	  sumv(i)=0.0d0
          sumqv(i)=0.0d0
	  sumtherm(i)=0.0d0
	enddo
	do i=1,nx
	do j=1,ny
	do k=1,nz
	  index=shifttab(dinx(i,j,k))
	  dinx(i,j,k)=index
	  dirtab(index)=dirtab(index)+1
	  freq=vz(i,j,k)/dny
	  sumv(index)=sumv(index)+freq
	  sumqv(index)=sumqv(index)+freq**2
	  freq=(vx(i,j,k)-vxshift)/dny
	  sumv(index)=sumv(index)+freq
	  sumqv(index)=sumqv(index)+freq**2
	  freq=(vy(i,j,k)-vyshift)/dny
	  sumv(index)=sumv(index)+freq
	  sumqv(index)=sumqv(index)+freq**2
	  sumtherm(index)=sumtherm(index)+vt(i,j,k)
	enddo
	enddo
	enddo
c	Fields with dirtab(i)=0 should not occur here
	do i=1,ntt
	    freqmom(i,1)=sumv(i)/(3*dirtab(i))
	    freqmom(i,2)=sqrt(2.0*(sumqv(i)/(3*dirtab(i))
     %        -freqmom(i,1)**2)+(sumtherm(i)/(dny*dirtab(i)))**2)
c	    freqmom(i,2)=sumtherm(i)/(dny*dirtab(i))
	enddo
	endif

	return
	end

c	*********************************************************
c	Make a table of 6 directions covering the distribution
c	given by its position and sigma
c	*********************************************************
	subroutine dirfromdst3(moments,n,table)
	implicit none
	integer n,l

	include 'fsizes.inc'
	real table(6),moments(ntab,6)

c	*********************************************************
c	Translate the table
c	*********************************************************
	do l=1,6
	  table(l)=moments(n,l)
	enddo

	return
	end

c	*********************************************************
c	Compute the corresponding density gradients for the
c	velocity gradients from the linear regression
c	*********************************************************
	subroutine bgrad(regcoeff,n,veltab,dgradtab)
	implicit none
	integer n,j,direct

	include 'fsizes.inc'
	parameter(direct=6)
	real regcoeff(ntab,6)
	real veltab(direct),dgradtab(direct)

c	*********************************************************
c	Translate the table
c	*********************************************************
	do j=1,6 
	  dgradtab(j)=regcoeff(n,j)/veltab(j)
c	  dgradtab(j)=0.0
	enddo
	return
	end

c	*********************************************************
c	Assign the gradient numbers according to their value
c	Largest gradients with highest index
c	*********************************************************
	subroutine gradorder(grad,order)
	implicit none

	real grad(6),xmin
	integer order(6),minnum,i,j,istor

c	Initialize order
	do i=1,6
	  order(i)=i
	enddo
c	Sort globally, simple bubblesort
	do i=1,5
	  xmin=grad(order(i))
	  minnum=i
	  do j=i+1,6
	   if (grad(order(j)).lt.xmin) then 
	    minnum=j
	    xmin=grad(order(j))
	   endif
	  enddo
c	Improve later - use only order
	  if (minnum.ne.i) then
	    istor=order(i)
	    order(i)=order(minnum)
	    order(minnum)=istor
	  endif
	enddo

	return
	end

c	********************************************************
c	Remove zero values as the gradients
c	********************************************************
	subroutine removezeros(sumf6,dirtab,ntt)
	implicit none
	integer ntt

	include 'fsizes.inc'
	real*8 sumf6(ntab,6),mx1,mx2,prod
	integer dirtab(ntab),i,j

	do i=1,ntt
c	Avoid empty table entries
	 if (dirtab(i).ne.0) then
	   prod=1.0d0
	   do j=1,6
	     prod=prod*sumf6(i,j)
	   enddo
c	   Only if there is a zero somewhere
	   if (prod.eq.0.0d0) then
	     mx1=min(sumf6(i,1),sumf6(i,3),sumf6(i,5))
	     mx2=max(sumf6(i,2),sumf6(i,4),sumf6(i,6))
c	     First pair
	     if (sumf6(i,1).eq.0.0d0) then
	       if (sumf6(i,2).eq.0.0d0) then
	         sumf6(i,1)=mx1*0.25
	         sumf6(i,2)=mx2*0.25
	       else
	         sumf6(i,1)=-sumf6(i,2)
	       endif
	     endif
	     if (sumf6(i,2).eq.0.0d0) sumf6(i,2)=-sumf6(i,1)
c	     Second pair
	     if (sumf6(i,3).eq.0.0d0) then
	       if (sumf6(i,4).eq.0.0d0) then
	         sumf6(i,3)=mx1*0.25
	         sumf6(i,4)=mx2*0.25
	       else
	         sumf6(i,3)=-sumf6(i,4)
	       endif
	     endif
	     if (sumf6(i,4).eq.0.0d0) sumf6(i,4)=-sumf6(i,3)
c	     Third pair
	     if (sumf6(i,5).eq.0.0d0) then
	       if (sumf6(i,6).eq.0.0d0) then
	         sumf6(i,5)=mx1*0.25
	         sumf6(i,6)=mx2*0.25
	       else
	         sumf6(i,5)=-sumf6(i,6)
	       endif
	     endif
	     if (sumf6(i,6).eq.0.0d0) sumf6(i,6)=-sumf6(i,5)
	   endif
	 endif
	enddo

	return
	end

