c	**************************************************************
c	Subroutine to compute the average radiative energy densities
c	seen by each density point depending on the local average
c	line profile
c	************************************************************
c	Translation intensity - energy density 
c	(J=2*pi/c*int(int(I(ny,z)*phi(ny,z),dny),dz)
c	Width of input distribution freqvar=sqrt(2)*sigma
c	************************************************************
	subroutine radexi(uout,uexi,freqmom,ntt,nff,lin)
	implicit none
	integer ntt,nff,lin

	include 'fsizes.inc'
	real uout(-nf:nf,ntra),uexi(ntab,ntra),freqmom(ntab,2)
	real prof,phi
	integer i,l,f

c	***********************************************************
c	Run for all table points (3 dimensions)
c	***********************************************************
	do i=1,ntt
	  do l=1,lin
	  uexi(i,l)=0.0
	  enddo
	 do f=-nff,nff
c	   Local profile
	   prof=phi(real(f),freqmom(i,1),freqmom(i,2))
	   do l=1,lin
	    uexi(i,l)=uexi(i,l)+prof*uout(f,l)
	   enddo
	 enddo
	enddo
	return
	end

c	**************************************************************
c	Small test routine to write out absorbed radiation field
c	**************************************************************
	subroutine printexi(i,j,inx,denstab,freqmom,uexi)
	implicit none
	integer i,j,direct

	include 'fsizes.inc'
	real denstab(ndens),freqmom(ntab,2)
	real uexi(ntab,ntra)
	integer inx(ns,ns,ns),index,k,l

	do k=1,128
	  index=inx(i,j,k)
	  write(6,*) i,j,k,denstab((index-1)/ngrad+1)
	  write(6,*) freqmom(index,1),freqmom(index,2)
	  write(6,*) uexi(index,1),uexi(index,2),uexi(index,3),
     %             uexi(index,4),uexi(index,5),uexi(index,6)
	enddo
	return
	end

c	************************************************************
c	Subroutines to carry out integration of the intensity 
c	convolved with the beam function for a given central beam
c	displacement
c	The background intensity is removed here.
c	**************************************************************
	subroutine intbeam(iout,ubg,factor,sigbeam,nx,ny,nff)
	implicit none
	integer nx,ny,nff

	include 'fsizes.inc'
	real iout(ns,ns,-nf:nf),iconv(ns,ns,-nf:nf)
	real sigbeam,ubg,factor,dist,sigsqu,spi,ftosg
	real wgt(-2*ns:2*ns,-2*ns:2*ns)
	real*8 dd(-nf:nf),norm

	integer i,j,k,ii,jj,idat,jdat,irange,istep,ni,nj

	real negexp,negint
	common /accgauss/ negexp,negint
	save /accgauss/

c	Normalization factor spi=1/pi
	data spi, ftosg / 0.318310, 1.66511/

c	**************************************************************
c	Translate intensity into beam temperature
c	**************************************************************
	do 100 i=1,nx
	do 100 j=1,ny
	do 100 k=-nff,nff
100	iout(i,j,k)=factor*(iout(i,j,k)-ubg)
c	**************************************************************
c	Pensil beams
c	**************************************************************
	sigsqu=sigbeam**2
	irange=int(negint*sigbeam)+1
	if (irange.lt.2) return
c	**************************************************************
c	Compute the field of weights within the beam
c	**************************************************************
	norm=0.0d0
	do 200 i=-irange,irange
	do 200 j=-irange,irange
	dist=sqrt(real(i**2+j**2))
	wgt(i,j)=spi/sigsqu*exp(-dist**2/sigsqu)
	norm=norm+wgt(i,j)
200	continue
c	**************************************************************
c	Consider spacing
c	In a future version half steps will be allowed too
c	**************************************************************
	istep=max(int(0.5*sigbeam*ftosg+0.1),1)
c	**************************************************************
c	Integration
c	**************************************************************
	ni=0
	do 300 i=1,nx,istep
	ni=ni+1
	nj=0
	do 300 j=1,ny,istep
	nj=nj+1
	do 350 k=-nff,nff
	dd(k)=0.0d0
350	continue
	do 400 ii=-irange,irange
	idat=mod(i+ii-1,nx)+1
	do 400 jj=-irange,irange
	jdat=mod(j+jj-1,ny)+1
	do 500 k=-nff,nff
	dd(k)=dd(k)+iout(idat,jdat,k)*wgt(ii,jj)
500	continue
400	continue
	do 450 k=-nff,nff
	iconv(ni,nj,k)=dd(k)/norm
450	continue
300	continue
c	**************************************************************
c	Reassign to the original field
c	**************************************************************
	nx=ni
	ny=nj
	do 800 i=1,nx
	do 800 j=1,ny
	do 800 k=-nff,nff
800	iout(i,j,k)=iconv(i,j,k)
	return
	end

c	**************************************************************
c	Subroutine translating observational to quantities
c	**************************************************************
	subroutine transobs(dny,fwhm,dist,deltar,sigbeam,off,kerr)
	implicit none
	real fwhm,dist,deltar,sigbeam,off,dny,clicht
	integer istep,kerr

	real astor,ftosg

	include 'fsizes.inc'
	real negexp,negint
	common /accgauss/ negexp,negint
	save /accgauss/

c	astor translates arcsec to rad
c	astor = pi/(180*60*60)
c	ftosg translates FWHM to sigma
c	ftosg = 2*sqrt(ln2)
	data astor,ftosg / 4.848137e-6,1.66511/
c       Speed of light in km/s
        data clicht/ 2.99792e5 /

	dny=dny/clicht
	sigbeam=astor*dist*fwhm/(ftosg*deltar)
c	Check whether the beam is too large for integration field
c	These two lines must agree with the expressions in intbeam
	if (int(negint*sigbeam)+1.gt.2*ns) kerr=1
	istep=max(int(0.5*sigbeam*ftosg+0.1),1)
	off=deltar*istep/(astor*dist)
	return
	end

c	**************************************************************
c	Subroutine translating the fractal fields into physical
c	quantities
c	**************************************************************
	subroutine transphys(dco,vx,vy,vz,vt,t,dh,nx,ny,nz)
	implicit none
	integer nx,ny,nz

	include 'fsizes.inc'
	real dco(ns,ns,ns),vz(ns,ns,ns),vt(ns,ns,ns)
	real vx(ns,ns,ns),vy(ns,ns,ns)
	real d0,ddd,ddv,tt,x,vturb,t,dh
	real clicht,cst,turb,vtherm
	integer i,j,k

	common /model/ d0,ddd,ddv,tt,x,vturb
	save /model/

c       Speed of light in km/s
        data clicht/ 2.99792e5 /
c       Translation FWHM-sigma cst=1/(2*sqrt(ln 2))
        data cst /0.600561/

c	**************************************************************
c	General translation
c	**************************************************************
	t=tt
	dh=1.0/x
	turb=sqrt(vtherm(t)+(cst*vturb)**2)/clicht
c	**************************************************************
c	Fields
c	**************************************************************
	do 100 i=1,nx
	do 100 j=1,ny
	do 100 k=1,nz
	vt(i,j,k)=turb
	vx(i,j,k)=vx(i,j,k)*ddv/clicht
	vy(i,j,k)=vy(i,j,k)*ddv/clicht
	vz(i,j,k)=vz(i,j,k)*ddv/clicht
	dco(i,j,k)=(dco(i,j,k)*ddd+d0)*x
c	dco(i,j,k)=(10.0**(dco(i,j,k)*ddd)*d0)*x
100	continue
	return
	end

c	**************************************************************
c	Subroutine to get the statistical properties of a 3D field
c	**************************************************************
	subroutine extremes3d(dco,nx,ny,nz,res,point)
	implicit none
	integer nx,ny,nz

	include 'fsizes.inc'
	real dco(ns,ns,2*nf+1),res(4)
	real dmax,dmin,wgt
	real*8 sumd,sumqd
	integer imax,jmax,kmax,point(3),i,j,k,k1,k2

c	*************************************************************
c	Either spatial z scale or frequency scale
c	*************************************************************
	if (nz.lt.0) then
	 k1=nf+1+nz
	 k2=nf+1-nz
	 wgt=real(nx*ny*(1-2*nz))
	else
	 k1=1
	 k2=nz
	 wgt=real(nx*ny*nz)
	endif
c	*************************************************************
c	Scan the fields
c	*************************************************************
	sumd=0.0d0
	sumqd=0.0d0
	dmin=1e30
	dmax=-1e30
	imax=1
	jmax=1
	kmax=1
	do 100 i=1,nx
	do 100 j=1,ny
	do 100 k=k1,k2
	sumd=sumd+dco(i,j,k)
	sumqd=sumqd+dco(i,j,k)**2
	dmin=min(dmin,dco(i,j,k))
	if (dco(i,j,k).gt.dmax) then
	  dmax=dco(i,j,k)
	  imax=i
	  jmax=j
	  kmax=k
	endif
100	continue
c	**************************************************************
c	Assign the results
c	**************************************************************
	res(1)=sumd/wgt
	res(2)=sqrt(sumqd/wgt-(sumd/wgt)**2)
	res(3)=dmin
	res(4)=dmax
	point(1)=imax
	point(2)=jmax
	point(3)=kmax
	return
	end

c	**************************************************************
c	Subroutine to get the statistical properties of a 2D field
c	**************************************************************
	subroutine extremes2d(dco,nx,ny,res,point)
	implicit none
	integer nx,ny

	include 'fsizes.inc'
	real dco(ns,ns),res(4)
	real dmax,dmin,wgt
	real*8 sumd,sumqd
	integer imax,jmax,point(3),i,j

c	*************************************************************
c	Scan the fields
c	*************************************************************
	sumd=0.0d0
	sumqd=0.0d0
	dmin=1e30
	dmax=-1e30
	imax=1
	jmax=1
	do 100 i=1,nx
	do 100 j=1,ny
	sumd=sumd+dco(i,j)
	sumqd=sumqd+dco(i,j)**2
	dmin=min(dmin,dco(i,j))
	if (dco(i,j).gt.dmax) then
	  dmax=dco(i,j)
	  imax=i
	  jmax=j
	endif
100	continue
c	**************************************************************
c	Assign the results
c	**************************************************************
	wgt=real(nx*ny)
	res(1)=sumd/wgt
	res(2)=sqrt(sumqd/wgt-(sumd/wgt)**2)
	res(3)=dmin
	res(4)=dmax
	point(1)=imax
	point(2)=jmax
	return
	end
